module Apply where

import RIO
import RIO.Process (HasProcessContext)
import qualified RIO.ByteString as B
import qualified RIO.ByteString.Lazy as LB
import qualified RIO.HashMap as H
import qualified RIO.Text as T
import IOClasses
import IOData
import Data
import Data.Binary (decodeOrFail)
import Data.Binary.Get (ByteOffset)
import CliUtils (getPeerUUID)
import RemoteUpdates
import Maildir
import Match (mailBoxIf)
import Uux (applyUux)

failedParsePeer :: Int
failedParsePeer = 61

peerParseFailure :: HasLogFunc env => String -> RIO env ()
peerParseFailure peer = do
  logError . display $ "Failed to parse peer: " `T.append` (T.pack $ show peer)
  exitWith $ ExitFailure failedParsePeer

applySingleCli :: (HasMailSyncConfig env, HasMailRoot env, HasMailDeliveryInfo env
                  , HasLogFunc env, HasProcessContext env)
               => PeerId -> RIO env ()
applySingleCli peer = do
--  let mPeer = UUID.fromString peerString
  -- maybe (peerParseFailure peerString) `flip` mPeer $ \peer ->
    withGlobalStateLocked $ \gs -> do
      input <- B.hGetContents stdin
      let (content :: Maybe Change) = readText input
      maybe (return ()) (\c -> void $ applyChange gs peer (changeId c) (changePatch c)) content

failedChangeParse :: Int
failedChangeParse = 58

changeParseFailure :: HasLogFunc env => (LB.ByteString, ByteOffset, String) -> RIO env ()
changeParseFailure (_, _, e) = do
  logError . display $ "Failed to parse change: " `T.append` (T.pack $ show e)
  exitWith $ ExitFailure failedChangeParse

applyManyBinary :: (HasMailSyncConfig env, HasMailRoot env, HasMailDeliveryInfo env
                   , HasLogFunc env, HasProcessContext env)
                => String -> RIO env ()
applyManyBinary ident = do
  mPeer <- getPeerUUID ident
  case mPeer of 
    Nothing -> return ()
    Just peer -> withGlobalStateLocked $ \gs -> do
      input <- LB.hGetContents stdin
      unless (LB.null input) $ applyManyBinary' peer input gs

applyManyBinary' :: (HasMailSyncConfig env, HasMailRoot env, HasMailDeliveryInfo env
                    , HasLogFunc env, HasProcessContext env)
                 => PeerId -> LB.ByteString -> GlobalState -> RIO env ()
applyManyBinary' peer input gs = do
  let parsed = decodeOrFail input
  either (changeParseFailure) `flip` parsed $ \(input', _, c) -> do
    gs' <- applyChange gs peer (changeId c) (changePatch c)
    logInfo . display $ "Applied change: " `T.append` (T.pack $ show c)
    unless (LB.null input') $ applyManyBinary' peer input' gs'


applyChange :: ( HasMailSyncConfig env, HasMailRoot env, HasMailDeliveryInfo env
               , HasLogFunc env, HasProcessContext env)
             => GlobalState -> PeerId -> PeerState -> ChangePayload -> RIO env GlobalState
applyChange gs peer peerState (ChangePatch patch) = applySingleGuarded gs peer peerState patch
applyChange gs peer peerState (ChangeUUX uux) = do
  applyUux peer (uuxCmd uux) (uuxStdin uux)
  writeGlobalState $ acceptIndexOnly peer peerState gs

applySingleGuarded :: ( HasMailSyncConfig env, HasMailRoot env, HasMailDeliveryInfo env
                      , HasLogFunc env )
                    => GlobalState -> PeerId -> PeerState -> Patch -> RIO env GlobalState
applySingleGuarded gs peer peerState patch = 
  let oldPeerState = fromMaybe (Id { fromId = 0}) . H.lookup peer $ upstreamPeers gs
      apply = applySingle' gs peer peerState patch
      nonApply = return $ acceptNonApply peer peerState gs patch
  in if peerState <= oldPeerState
     then do
       logWarn . display . T.pack $ "Ignoring out-of-order pach " ++ show peerState ++ " from " ++ show peer
       return gs
     else mailBoxIf (fromFSName . fromMailbox $ mailbox patch) apply nonApply


applySingle' :: ( HasMailSyncConfig env, HasMailRoot env, HasMailDeliveryInfo env
                , HasLogFunc env )
             => GlobalState -> PeerId -> PeerState -> Patch -> RIO env GlobalState
applySingle' gs peer peerState patch = do
  gs' <- case operation patch of
       Add content flags -> do
         lname <- newMail (mailbox patch) content flags
         return $ acceptNewMail peer peerState gs (mailbox patch) (globalName patch) lname content flags
       Delete -> do
         let mOnDisk = myBase gs (mailbox patch) (globalName patch)
         maybe (warnPatch peer patch) (void . deleteMail (mailbox patch)) mOnDisk
         return $ acceptMailRemoval peer peerState gs (mailbox patch) (globalName patch)
       Flags newFlags -> do 
         let mOnDisk = myBase gs (mailbox patch) (globalName patch)
         maybe (warnPatch peer patch) (void . changeFlags (mailbox patch) newFlags) mOnDisk
         return $ acceptFlagChange peer peerState gs (mailbox patch) (globalName patch) newFlags
  writeGlobalState gs'

warnPatch :: (HasLogFunc env) => PeerId -> Patch -> RIO env ()
warnPatch peer patch = logWarn . display $ "Failed to apply patch: " 
                       `T.append` (T.pack $ show patch) `T.append` " from: " `T.append`
                       (T.pack $ show peer)
