module InitialSync ( initialSyncForPeer
                   , initialSyncFrom
)
where

import RIO
import Data.Binary (encode, decodeOrFail)
import Data.Binary.Get (ByteOffset)

import qualified RIO.ByteString.Lazy as LB
import qualified RIO.ByteString as BS
import qualified RIO.HashMap as H
import qualified RIO.Set as S
import qualified RIO.Text as T

import qualified ForeignUtils as FF

import Data
import IOClasses
import IOData ( withGlobalStateLocked
              , writeGlobalState
              , writeGlobalState_
              , writePeerState
              , addPeer
              )
import LocalUpdates (getMails, removeLostEntries)
import Maildir (findMail, newMailQD)
import RemoteUpdates (acceptPeer, anyPeer)
import Utils (maybeReadBinaryFile, (>>=?))

initialSyncForPeer :: ( HasMailSyncConfig env
                      , HasMailRoot env
                      , HasLogFunc env
                      , HasOptions env
                      )
                      => PeerId -> String -> RIO env ()
initialSyncForPeer peer name = withGlobalStateLocked $ \origGs -> do
  let mailboxes = H.keys $ state origGs
  origGs' <- foldM `flip` origGs `flip` mailboxes $ \st mb -> do
    let mails = fromMaybe S.empty . H.lookup mb $ state st
    foldM `flip` st `flip` mails $ \ gs mail -> gs `seq` do
      mbContent <- findMail mb (toOnDiskMail mail) >>=? maybeReadBinaryFile
      maybe (writeGlobalState . removeLostEntries gs mb . S.singleton $ mglobalName mail)
         `flip` mbContent
         $ \ content -> do
               let toSend = InitialMail $ InitialSyncMail { iMailBox = mb
                                                          , iGlobalName = mglobalName mail
                                                          , iFlags = mFlags mail
                                                          , iBody = content
                                                          }
               BS.hPut stdout . toStrictBytes $ encode toSend
               return gs
  let peerState = Id { fromId = (fromId $ nextId origGs') -1 }
  let doneMessage = InitialDoneSync 
                    $ InitialSyncState 
                      { iMyId = myUUID origGs'
                      , iMyName = myName origGs'
                      , iYouKnowUntil = peerState
                      }
  let gs' = acceptPeer origGs' peer
  writeGlobalState_ gs'
  writePeerState peer peerState
  addPeer (name, peer)
  BS.hPut stdout . toStrictBytes $ encode doneMessage

failedInitialParse :: Int
failedInitialParse = 57

initialParseFailure :: HasLogFunc env 
                    => (LB.ByteString, ByteOffset, String) -> RIO env ()
initialParseFailure (_, _, e) = do
  logError . display $ "Failed to parse initial message: " `T.append` (T.pack $ show e)
  exitWith $ ExitFailure failedInitialParse

initialSyncFrom :: ( HasMailSyncConfig env
                   , HasLogFunc env
                   , HasMailRoot env
                   , HasMailDeliveryInfo env
                   , HasOptions env
                   )
                => RIO env ()
initialSyncFrom = withGlobalStateLocked $ \gs ->
  unless (anyPeer gs) $ LB.hGetContents stdin >>= initialSyncFrom' gs

initialSyncFrom' :: ( HasMailSyncConfig env
                    , HasLogFunc env
                    , HasMailRoot env
                    , HasMailDeliveryInfo env
                    , HasOptions env
                    )
                 => GlobalState -> LB.ByteString -> RIO env ()
initialSyncFrom' gs input = do
  let parsed = decodeOrFail input
  either initialParseFailure `flip` parsed $ \(input', _, msg) -> do
    case msg of 
      InitialDoneSync s  -> do
        FF.sync
        writeGlobalState_ (gs { upstreamPeers = H.singleton (iMyId s) (iYouKnowUntil s) })
        addPeer (iMyName s, iMyId s)
      InitialMail m -> do
        let mb = iMailBox m
        baseName <- newMailQD mb (iBody m) (iFlags m)
        let mail = Mail { mBase = baseName
                        , mglobalName = iGlobalName m
                        , mFlags = iFlags m
                        }
        let gs' = gs { state = H.insert mb (S.insert mail $ getMails mb gs) $ state gs }
        initialSyncFrom' gs' input'
